#!/bin/bash

#set -x

ORIGIN_DIR="../main/"
WORK_DIR="./build/"
TARGET_DIR="./source/"

PHASE=${1:-ALL}
FILES=${2:-$(cat destfiles.txt | grep -v ^#)}

for destfile in $FILES; do
    file=`echo $destfile | sed 's#.*/##g'`
    chapterdir=`echo $destfile | sed 's#/.*##g'`
    filebase=`basename $destfile .tex`

    mkdir -p ${WORK_DIR}${chapterdir}
    mkdir -p ${TARGET_DIR}${chapterdir}

    if [ $PHASE == "ALL" ] || [ $PHASE == "prepandoc" ]; then
        printf "%-50s (%s)\n" "$destfile" "prepandoc"
        ./pre_conversion_changes.sh  ${ORIGIN_DIR}${file} ${WORK_DIR}${destfile}
    fi

    if [ $PHASE == "ALL" ] || [ $PHASE == "pandoc" ]; then
        printf "%-50s (%s)\n" "$destfile" "pandoc"
        pandoc --verbose --columns=500  -f latex+raw_tex -t rst ${WORK_DIR}${destfile} -o ${WORK_DIR}${chapterdir}/${filebase}.rst || exit "could not convert file ${WORK_DIR}${destfile}"
        #pandoc --verbose --columns=500  -f latex+raw_tex -t rst --filter latex-scan.py ${TARGET_DIR}${destfile} -o ${TARGET_DIR}${chapterdir}/${filebase}.rst || exit "could not convert file $file"
    fi

    if [ $PHASE == "ALL" ] || [ $PHASE == "postpandoc" ]; then
        printf "%-50s (%s)\n" "$destfile" "postpandoc"
        #    ./post_conversion_changes.sh ${TARGET_DIR}${chapterdir}/${filebase}.rst
        printf "%s\n"   ".. ATTENTION do not edit this file manually." > ${TARGET_DIR}${chapterdir}/${filebase}.rst
        printf "%s\n\n" "   It was automatically converted from the corresponding .tex file" >> ${TARGET_DIR}${chapterdir}/${filebase}.rst
        cat ${WORK_DIR}${chapterdir}/${filebase}.rst | ./latex-scan.py --standalone >> ${TARGET_DIR}${chapterdir}/${filebase}.rst 2>${WORK_DIR}${chapterdir}/${filebase}.latex-scan.log || exit "failed to post convert ${WORK_DIR}${chapterdir}/${filebase}.rst"
    fi
done
